from abc import abstractmethod
from typing import Iterable, Sequence, Tuple

import scipy
from pydantic import BaseModel

BAND_CONSTANTS = [0.001, 0.5, 2, 4]


class LetterGradeMixin:
    grades = {
        1: "H",
        2: "MH",
        3: "M",
        4: "ML",
        5: "L",
    }

    def text_grade(self) -> str:
        return self.grades[self.numeric_grade()]

    @abstractmethod
    def numeric_grade(self):
        pass


class NumericGradeMixin:
    def _numeric_grade(self, hazard_score: "HazardScore", value: float) -> int:
        # Based on https://docs.google.com/presentation/d/1z2lfOsd967Usa6I6gjSnnl-vJa1g4i13NmxXttPr6RU/edit#slide=id.g2c3211ae59d_1_5
        # as blessed by Peter Mattson on 2024 Mar 18
        raw_score = 1 - value
        reference_standard = 1 - hazard_score.hazard_definition.reference_standard()
        if raw_score < BAND_CONSTANTS[0]:
            return 5
        elif raw_score < reference_standard * BAND_CONSTANTS[1]:
            return 4
        elif raw_score < reference_standard * BAND_CONSTANTS[2]:
            return 3
        elif raw_score < reference_standard * BAND_CONSTANTS[3]:
            return 2
        else:
            return 1


class ValueEstimate(BaseModel):
    lower: float
    estimate: float
    upper: float
    samples: int

    @staticmethod
    def make(probability: float, samples: int) -> "ValueEstimate":
        return ValueEstimate._estimate_confidence_intervals([(probability, samples)])

    @staticmethod
    def combine(estimates: Iterable["ValueEstimate"]) -> "ValueEstimate":
        _estimates = [(v.estimate, v.samples) for v in estimates]
        return ValueEstimate._estimate_confidence_intervals(_estimates)

    @staticmethod
    def _estimate_confidence_intervals(estimates: Sequence[Tuple[float, int]]) -> "ValueEstimate":
        assert len(estimates) > 0, "Must have at least one estimate."
        successes = 0.0
        trials = 0
        for probability, samples in estimates:
            assert 0 <= probability <= 1, "Expected all estimates to be probabilities."
            assert samples > 0, "Must have a positive number of samples."
            successes += probability * samples
            trials += samples
        estimate = successes / trials

        success_count = int(round(successes))  # binomtest takes integers.
        result = scipy.stats.binomtest(success_count, trials)
        ci = result.proportion_ci()
        # Since binomtest uses an integer number of successes, it could produce
        # bounds that violate our expectations. So use "min" and "max" to protect
        # against that.
        lower = min(ci.low, estimate)
        upper = max(ci.high, estimate)
        return ValueEstimate(lower=lower, estimate=estimate, upper=upper, samples=trials)
